#!/usr/bin/perl -w

# UpStare is a system that can apply immediate dynamic software updates
# in multi-threaded userspace applications using stack
# reconstruction. It is Copyright (C) 2006 by Kristis Makris
# <kristis.makris@asu.edu,mkgnu@mkgnu.net>.

# In executable binary form produced and digitally signed specifically
# by Kristis Makris, it is freely distributed for personal and
# academic/research use. It is NOT distributed for commercial use. No
# other binary forms can be distributed.

# The source code and documentation, either in electronic or hardcopy
# format, may not be used for any purpose whatsoever without the written
# permission of Kristis Makris.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


use strict;
use Data::Dumper;



my $installation_dir = '/usr/include';
my $cilly = '/usr/bin/hcu_cilly.sh';
my $runtime_headers = '/usr/include/upstare/runtime';

my $hcu_transform = "$cilly -I $runtime_headers -Wno-attributes --save-temps --stats --noPrintLn --keepunused";
my $hcu_runtime_flags = "-ldl -pthread -lhcu ";
my $hcu_flags = "-ldl";



main ( @ARGV );



sub main {
    my ( @arguments ) = ( @_ );

    if (scalar @arguments <= 0) {
	print "usage: $0 <sourcefile> <flags>\n";
	exit -1;
    }

    # We NEED to prepend header definitions to the source file
    # first. It seems that CIL preprocesses source without giving a
    # module writer the opportunity to prepend header '#include's and
    # then evaluate the #includes and macros. This can lead to
    # duplicate '#include's if a CIL module attempted to insert an
    # include of an already '#include'd header (e.g. pthreads.h)
    my $source_files = determine_source_files( @arguments );
    my $hcu_new_args = join ( " ", @arguments );
    foreach my $hcu_source_file ( sort keys %{ $source_files } ) {
	my $hcu_new_source_file = $hcu_source_file . ".hcucc.c";
        # If the source file contains any '+' in it, escape them
        my $escaped_hcu_source_file = $hcu_source_file;
        $escaped_hcu_source_file =~ s/\+/\\\+/g;
        # Any dots in the escaped source file should become escaped,
        # else the dot in the regular expression will be interpreted
        # as "any character"
        $escaped_hcu_source_file =~ s/\./\\./g;
        #print "source_file: '$hcu_source_file'\n";
        #print "new_source_file: '$hcu_new_source_file'\n";
        #print "escaped_source_file: '$escaped_hcu_source_file'\n";
	$hcu_new_args =~ s/$escaped_hcu_source_file/$hcu_new_source_file/;
	append_after_last_include( $hcu_source_file, $hcu_new_source_file );

        # If this is a source file that is a configuration file, we
        # need to preprocess it before we let a CIL module parse it
        # internally
        if ( $source_files->{ $hcu_source_file } eq "preprocess") {
            my $hcu_preprocessed_source_file = $hcu_source_file . ".hcudiff.c";
            my $cmd = "$hcu_transform -E $hcu_new_source_file -o $hcu_preprocessed_source_file";
            system ( $cmd );
            my $system_retval = $?;
            handle_retval( $system_retval );
            $hcu_new_args =~ s/$hcu_new_source_file/$hcu_preprocessed_source_file/;
        }
    }

    my $cmd = "$hcu_transform $hcu_flags $hcu_new_args $hcu_runtime_flags";

    # Try to escape double quotes (") that are not already escaped
    # with a '\'. Otherwise the arguments supplied to the command-line
    # will not be passed as originally intented: the single quotes (')
    # will be dropped. For example, an argument like:
    #
    #      -DSYSCONFDIR="/usr/local/pgsql/etc"
    #
    # will be passed for execution as
    #
    #      -DSYSCONFDIR=/usr/local/pgsql/etc
    #
    # which will cause the variable to no longer hold a string. But
    # Don't escape the double quote if it is already escaped, like:
    #
    # -DPKGLIBDIR=\"/tmp/postgresbuild/lib\"
    #
    # since this will be passed for execution as
    #
    # -DPKGLIBDIR="/tmp/postgresbuild/lib"
    #
    # which is what we want
    $cmd =~ s/([^\\])\"/$1\\\"/g;

    # Run the command
    # 
    # But make sure that warnings related to the attributes accepted
    # by our compiler are not reported
    system ( $cmd );
    my $system_retval = $?;
    handle_retval( $system_retval );

    # After compilation, move the object file to have the expected
    # output filename (.o, not .c.hcucc.o)
    foreach my $hcu_source_file ( sort keys %{ $source_files } ) {
	my $hcu_old_output_file = $hcu_source_file . ".hcucc.o";
	my $hcu_new_output_file = $hcu_source_file;
	$hcu_new_output_file =~ s/\.c$/\.o/;
	#print "Will rename '$hcu_old_output_file' to '$hcu_new_output_file'\n";
	rename $hcu_old_output_file, $hcu_new_output_file;

        # Also, remove the source files that contained our prepended
        # header definitions.
        #print "Will unlink '$hcu_source_file.hcucc.c'\n";
        #unlink $hcu_source_file . ".hcucc.c";
    }
}



sub determine_source_files {
    my ( @arguments ) = ( @_ );
    my $arguments_string = join ( " ", @arguments );
    my $source_files;

    # Examine each argument
    foreach my $argument ( @arguments ) {
        
        if ( $argument =~ m/(\S+)\.c\s+/ ||
             $argument =~ m/(\S+)\.c$/ ) {
            # If it is a C source file
            my $match_file = $1;
            if ( $match_file =~ m/--compiler-configuration-file=(.*)/ ) {
                # If it is a compiler configuration file, treat it specially
                my $source_file = $1 . ".c";
                $source_files->{ $source_file } = "preprocess";
            } else {
                # It is the actual C source file that should be compiled
                my $source_file = $match_file . ".c";
                $source_files->{ $source_file } = "rename";
            }
        }
    }

    return $source_files;
}



sub append_after_last_include {
    my ( $hcu_source_file, $hcu_new_source_file ) = ( @_ );

    my $content = file_content ( $hcu_source_file );
    my $new_content = $content;

    # Prepend our headers
    $new_content = '#include "hcu_headers.h"' . "\n" . $new_content ;

    open OUTFILE, "> $hcu_new_source_file";
    print OUTFILE $new_content;
    close OUTFILE;
}


#
# Returns the contents of a file
#
# PARAMETERS:
# $1 - Name of file whose contents will be read
#
# RETURNS:
# Contents of file
sub file_content {
    my ( $filename ) = ( @_ );
    my $line;
    my $contents = "";

    if ( -e $filename ) {
	open ( FILE_CONTENT, "$filename" ) || die "File \"$filename\" cannot be opened\n";
	while ( $line = <FILE_CONTENT> ) {
	    $contents .= $line;
	}
	close ( FILE_CONTENT );
    } else {
	print "$0: File '$filename' does not exist\n";
	exit -1;
    }

    return $contents;
}


sub handle_retval {
    my ( $system_retval ) = ( @_ );

    if ($system_retval == -1) {
        print "$0: failed to execute: $!\n";
        exit -1;
    } elsif ($system_retval & 127) {
        printf "$0: child died with signal %d, %s coredump\n",
        ($system_retval & 127),  ($system_retval & 128) ? 'with' : 'without';
        exit -1;
    } elsif ($system_retval != 0) {
        printf "$0: child exited with value %d\n", $system_retval >> 8;
        exit -1;
    } else {
        # Execution succeeded.
    }
}
