/*
UpStare is a system that can apply immediate dynamic software updates
in multi-threaded userspace applications using stack
reconstruction. It is Copyright (C) 2006 by Kristis Makris
<kristis.makris@asu.edu,mkgnu@mkgnu.net>.

In executable binary form produced and digitally signed specifically
by Kristis Makris, it is freely distributed for personal and
academic/research use. It is NOT distributed for commercial use. No
other binary forms can be distributed.

The source code and documentation, either in electronic or hardcopy
format, may not be used for any purpose whatsoever without the written
permission of Kristis Makris.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef HCU_HEADER_MACROS_H
#define HCU_HEADER_MACROS_H



/** Maximum size of the name of each source programs' function */
#define HCU_MAX_FUNCTION_NAME_SIZE 256



/** Attribute used to identify functions that should not be processed
    for stack-reconstruction. */
#define HCU_NON_UPDATEABLE __attribute__((HCU_ATTRIBUTE_NON_UPDATEABLE))



/** Attribute used to identify functions calls which should be
    processed for stack-reconstruction even though they don't provide
    a definition. Those functions are usually declarations of
    functions in the original program source, which are called from
    the updated file. */
#define HCU_NON_LIBRARY_FUNCTION __attribute__((HCU_ATTRIBUTE_NON_LIBRARY_FUNCTION))



/** Attribute used to identify symbols whose version number should be
    set by stack reconstruction, as requested by the patch
    generator. */
#define HCU_VERSION_SHOULD_BE_SET __attribute__((HCU_ATTRIBUTE_VERSION_SHOULD_BE_SET))



/** Attribute used to identify symbols whose version number has
    already been set from the patch generator and should not be set
    again by the stack reconstruction. */
#define HCU_VERSION_HAS_BEEN_SET __attribute__((HCU_ATTRIBUTE_VERSION_HAS_BEEN_SET))



/** Attribute used to identify functions that are new versions of the
    originals. These functions will not have an indirection pointer
    created for them, since one was already created for the
    original. */
#define HCU_NEW_VERSION __attribute__((HCU_ATTRIBUTE_NEW_VERSION))



/** Attribute used to identify functions that will not be called
    commonly. They will be called when the stack is reconstructed. */
#ifdef __APPLE__
#define __HCU_ATTRIBUTE_UNCOMMON_CODE__ __section__(".hcu.uncommon,")
#else
#define __HCU_ATTRIBUTE_UNCOMMON_CODE__ __section__(".hcu.uncommon")
#endif



/** Macros used to improve the cache locality of uncommon code by
    placing it in a separate memory region. */
#define hcu_better_cache_locality_pre_simple(value,name) if (value) {
#define hcu_better_cache_locality_pre_deep_simple(value,name_if,name_else) if (value) {
#define hcu_better_cache_locality_post_deep_simple(name) }
#define hcu_better_cache_locality_post_simple() }

#define HCU_BETTER_CACHE_LOCALITY 0
#if HCU_BETTER_CACHE_LOCALITY
#define hcu_better_cache_locality_pre(value, name) __asm__ __volatile__ ( "cmpl $0x1,%0\n\t" \
                                                                   "je 1f\n\t" \
								   ".subsection 2\n\t" \
                                                                   ".section .hcu.uncommon_code\n" \
							           "hcu_uncommon_section_" name ":\n\t" \
							           "1:\n\t" \
                                                                   : /* no output */ \
					                           : "g" (value) )


#define hcu_better_cache_locality_pre_deep(value, name_if, name_else) __asm__ __volatile__ ( "cmpl $0x1,%0\n\t" \
                                                                   "je 1f\n\t" \
                                                                   "jmp " "hcu_uncommon_section_" name_else "\n\t" \
								   ".subsection 2\n\t" \
                                                                   ".section .hcu.uncommon_code\n" \
							           "hcu_uncommon_section_" name_if ":\n\t" \
							           "1:\n\t" \
                                                                   : /* no output */ \
					                           : "g" (value) )


#define hcu_better_cache_locality_post_deep(name) __asm__ __volatile__ ( "hcu_uncommon_section_" name ":\n\t" );
#define hcu_better_cache_locality_post() __asm__ __volatile__ ( "jmp 1f\n\t" \
                                                                ".text\n\t" \
                                                                "1:\n" );

#else
#define hcu_better_cache_locality_pre(value,name) hcu_better_cache_locality_pre_simple(value,name)
#define hcu_better_cache_locality_pre_deep(value,name_if,name_else) hcu_better_cache_locality_pre_deep_simple(value,name_if,name_else)
#define hcu_better_cache_locality_post_deep(name) hcu_better_cache_locality_post_deep_simple(name)
#define hcu_better_cache_locality_post() hcu_better_cache_locality_post_simple()
#endif /* HCU_BETTER_CACHE_LOCALITY */


#define hcu_simulate() __asm__ __volatile__ ( "nop\n\t" \
					      "nop\n\t" \
					      "nop\n\t" \
					      "nop\n\t" );



/** Macro used to statically define points where a dynamic software
    update should be executed without the assistance of an update
    thread.

    Applies for the updating model of immediate updates
*/
#define HCU_STATIC_REQUEST_UPDATE_IMMEDIATE() hcu_static_request_update(HCU_UPDATE_MODEL_IMMEDIATE)



/** Macro used to statically define points where a dynamic software
    update should be executed without the assistance of an update
    thread. 

    Applies for the updating model of lazy updates
*/
#define HCU_STATIC_REQUEST_UPDATE_LAZY() hcu_static_request_update(HCU_UPDATE_MODEL_LAZY)



/** Macro used to describe the update file that will be used for a
    static update. */
#define HCU_STATIC_UPDATE_FILE(filename, version) \
char hcu_static_update_filename[] = filename; \
long hcu_static_update_version = version;



/** We support updating multi-threaded programs. */
#define HCU_SUPPORT_MULTITHREADED_UPDATES 1


#if HCU_SUPPORT_MULTITHREADED_UPDATES
/* --------------------------------------------------------------------------*/

/** Macro that attempts to update a multi-threaded application */
#define hcu_try_to_update(from_function, local_stack_variable, local_stack_variable_size, local_stack_frame_fields, continuation_value, update_point_number) \
hcu_try_to_update_top_multithreaded(from_function, local_stack_variable, local_stack_variable_size, local_stack_frame_fields, continuation_value, update_point_number)

/** We support updating the bottom part of multi-threaded programs. */
#define hcu_try_to_update_bottom() hcu_try_to_update_bottom_multithreaded()

/** Macro that attempts to unwind the stack of a multi-threaded
    application */
#define hcu_must_unwind_up(from_function) hcu_must_unwind_up_multithreaded(from_function)


#else
/* --------------------------------------------------------------------------*/

/** Macro that attempts to update a single-threaded application. */
#define hcu_try_to_update(from_function, local_stack_variable, local_stack_variable_size, local_stack_frame_fields, continuation_value, update_point_number) \
hcu_try_to_update_top_singlethreaded(from_function, local_stack_variable, local_stack_variable_size, local_stack_frame_fields, continuation_value, update_point_number)


/** We support updating the bottom part of single-threaded programs. */
#define hcu_try_to_update_bottom() hcu_try_to_update_bottom_singlethreaded()

/** Macro that attempts to unwind the stack of a single-threaded
    application */
#define hcu_must_unwind_up(from_function) hcu_must_unwind_up_singlethreaded(from_function)



/* --------------------------------------------------------------------------*/
#endif



#endif /* HCU_HEADER_MACROS_H */

