/*
UpStare is a system that can apply immediate dynamic software updates
in multi-threaded userspace applications using stack
reconstruction. It is Copyright (C) 2006 by Kristis Makris
<kristis.makris@asu.edu,mkgnu@mkgnu.net>.

In executable binary form produced and digitally signed specifically
by Kristis Makris, it is freely distributed for personal and
academic/research use. It is NOT distributed for commercial use. No
other binary forms can be distributed.

The source code and documentation, either in electronic or hardcopy
format, may not be used for any purpose whatsoever without the written
permission of Kristis Makris.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef HCU_THREAD_LIST_H
#define HCU_THREAD_LIST_H



#include "hcu_update_points.h"
#include "hcu_lock.h"
#include "hcu_stack_trace.h"
#include <pthread.h>
#include <semaphore.h>



#define HCU_HASH_TABLE_SIZE 512



#define hcu_table_makekey(id) (id % HCU_HASH_TABLE_SIZE)



typedef struct {
  long port;
  int sockfd;
  char *hostname;
  int blocked_on_update;
  int blocked_on_stack_unwind_up;
  int blocked_on_stack_reconstruction;
} hcu_thread_list_item_children_t;

typedef struct {
  hcu_stack_trace_t trace;
} hcu_thread_list_item_stack_traces_t;

typedef struct {
  pthread_mutex_t *lock;
} hcu_thread_list_item_want_t ;

typedef struct {
  pthread_mutex_t *lock;
} hcu_thread_list_item_have_t ;

typedef struct {
} hcu_thread_list_item_blocked_t ;

typedef struct {
} hcu_thread_list_item_blocked_on_stack_unwind_up_t ;

typedef struct {
} hcu_thread_list_item_blocked_on_stack_reconstruction_t ;

typedef struct {
  char *thread_description;
  int must_reconstruct_callee_stack;
  /** Pointer to the name of the function up to which the stack should
      unwind before stack reconstruction downwards will begin again. */
  char *requested_stack_top;
} hcu_thread_list_item_description_t ;

typedef struct {
  /** Pointer to an array of update points that can be visited by this
      thread. */
  hcu_update_point_t *update_points;
} hcu_thread_list_item_update_points_t ;

typedef union {
  hcu_thread_list_item_children_t children;
  hcu_thread_list_item_stack_traces_t stack_traces;
  hcu_thread_list_item_have_t want;
  hcu_thread_list_item_have_t have;
  hcu_thread_list_item_blocked_t blocked;
  hcu_thread_list_item_blocked_on_stack_unwind_up_t blocked_on_stack_unwind_up;
  hcu_thread_list_item_blocked_on_stack_reconstruction_t blocked_on_stack_reconstruction;
  hcu_thread_list_item_description_t description;
  hcu_thread_list_item_update_points_t update_points;
} hcu_thread_list_item_task_t;

typedef enum {
  HCU_THREAD_LIST_ITEM_TASK_TYPE_CHILDREN = 1,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_STACK_TRACES = 2,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_WANT = 3,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_HAVE = 4,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_BLOCKED = 5,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_BLOCKED_ON_STACK_UNWIND_UP = 6,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_BLOCKED_ON_STACK_RECONSTRUCTION = 7,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_DESCRIPTION = 8,
  HCU_THREAD_LIST_ITEM_TASK_TYPE_UPDATE_POINTS = 9
} hcu_thread_list_item_task_type_t;


/* The id can be either a thread id (pthread_t) or a process id
   (pid_t). We use the largest possible datatype that can accomodate
   both of them. */
typedef unsigned long hcu_thread_or_child_id_t;

/* This data structure is used for a wide range of tasks. This is
   still not the most efficient way to implement it. */
typedef struct{
  hcu_thread_or_child_id_t id;
  short task_type;
  hcu_thread_list_item_task_t task;
} hcu_thread_list_item_t;

struct hcu_thread_list_s {
  hcu_thread_list_item_t *item;
  struct hcu_thread_list_s *next;
};
typedef struct hcu_thread_list_s hcu_thread_list_t;

typedef struct {
  hcu_thread_list_t *list;
} hcu_table_item_t;

typedef struct {
  hcu_lock_t lock;
  hcu_table_item_t table[HCU_HASH_TABLE_SIZE];
} hcu_table_t;


hcu_thread_list_item_t *hcu_thread_list_item_create(hcu_thread_list_item_task_type_t task_type);
void hcu_thread_list_item_destroy(hcu_thread_list_item_t *item);
int hcu_thread_list_item_equal(hcu_thread_list_item_t *item1, hcu_thread_list_item_t *item2);

hcu_thread_list_t *hcu_thread_list_create();
void hcu_thread_list_destroy(hcu_thread_list_t *list);
void hcu_thread_list_add(hcu_thread_list_t **list, hcu_thread_list_item_t *item);
hcu_thread_list_t *hcu_thread_list_remove(hcu_thread_list_t **list, hcu_thread_list_item_t *removed_item);
hcu_thread_list_t *hcu_thread_list_remove_id(hcu_thread_list_t **list, hcu_thread_or_child_id_t id);
hcu_thread_list_t *hcu_thread_list_lookup_lock(hcu_thread_list_t *list, pthread_mutex_t *lock);
hcu_thread_list_t *hcu_thread_list_lookup_id(hcu_thread_list_t *list, hcu_thread_or_child_id_t id);
hcu_thread_list_t *hcu_thread_list_lookup_thread_description(hcu_thread_list_t *list, char *thread_description);
long hcu_thread_list_count(hcu_thread_list_t *list);
char * hcu_thread_list_tostring(hcu_thread_list_t *list);
char * hcu_thread_list_update_points_tostring(hcu_thread_list_t *list);

hcu_table_t *hcu_table_create();
void hcu_table_destroy(hcu_table_t **table);
void hcu_table_add(hcu_table_t *table, hcu_thread_list_item_t *item);
hcu_thread_list_t *hcu_table_remove(hcu_table_t *table, hcu_thread_list_item_t *item);
hcu_thread_list_t *hcu_table_remove_id(hcu_table_t *table, hcu_thread_or_child_id_t id);
hcu_thread_list_t *hcu_table_lookup_lock(hcu_table_t *table, hcu_thread_or_child_id_t id, pthread_mutex_t *lock);
hcu_thread_list_t *hcu_table_lookup_id(hcu_table_t *table, hcu_thread_or_child_id_t id);
long hcu_table_count(hcu_table_t *table);
void hcu_table_lock(hcu_table_t *table);
void hcu_table_unlock(hcu_table_t *table);
char * hcu_table_tostring(hcu_table_t *table);
char * hcu_table_update_points_tostring(hcu_table_t *table);
void hcu_block_indefinitely();



#endif /* THREAD_LIST_H */
